#!/usr/bin/python
# -*- coding: utf-8 -*-

# modules from the std-library
import csv
import os
import re
import socket
import subprocess
import sys
import types
import xmlrpclib
from optparse import OptionParser

# currently we only support linux
if not sys.platform.startswith('linux'):
	print("ERROR: %s not supported!" % sys.platform)
	sys.exit(-1)

# modules in bin-path
import enum

# external libs
# python-lxml module
try:
	from lxml import etree
except ImportError:
	print("python-lxml module not found! (python-lxml)")
	print("see http://codespeak.net/lxml/")
	print("programm terminating ...!")
	sys.exit(-1)

# statistics module
try:
	import pstat
except ImportError:
	print("pstat module not found! (python-stats)")
	print("see http://www.nmr.mgh.harvard.edu/Neural_Systems_Group/gary/python.html")
	print("programm terminating ...!")
	sys.exit(-1)

# pyparsing module
try:
	import pyparsing as pypa
	pypa.ParserElement.enablePackrat()		# speed up parsing
	sys.setrecursionlimit(2000)				# handle larger expressions
except ImportError:
	print("pyparsing module not found! (python-pyparsing)")
	print("see http://pyparsing.wikispaces.com/")
	print("programm terminating ...!")
	sys.exit(-1)

##################################################
# config:
__outputfile = "cppstats.csv"
__outputfexp = "cppfeatureexpressions.txt"

# error numbers:
__errorfexp = 0
__errormatch = []
##################################################


##################################################
# constants:
# namespace-constant for src2srcml
__cppnscpp = 'http://www.sdml.info/srcML/cpp'
__cppnsdef = 'http://www.sdml.info/srcML/src'
__cpprens = re.compile('{(.+)}(.+)')

# conditionals - necessary for parsing the right tags
__conditionals = ['if', 'ifdef', 'ifndef']
__conditionals_elif = ['elif']
__conditionals_else = ['else']
__conditionals_endif = ['endif']
__conditionals_all = __conditionals + __conditionals_elif + \
		__conditionals_else
__macro_define = ['define']
__macrofuncs = {}		# functional macros like: "GLIBVERSION(2,3,4)",
						# used as "GLIBVERSION(x,y,z) 100*x+10*y+z"
__curfile = ''			# current processed xml-file
__defset = set()		# macro-objects
__defsetf = dict()		# macro-objects per file

# collected statistics
__statsorder = enum.Enum(
	'FILENAME',			# name of the file
	'LOC',				# lines of code
	'NOFC',				# number of feature constants
	'NOFE',				# number of features expressions
	'NOFE1',			# number of features expressions on level 1
	'LOF',				# number of feature code lines
	'LOFMIN',			# minimum number of lof (single feature)
	'LOFMAX',			# maximum number of lof (single feature)
	'LOFMEAN',			# average number of lof (single feature)
	'LOFSTD',			# standard-deviation of lof (single feature)
	'NNIMAX',			# number of nested ifdefs
	'NNIMEAN',			# average nested ifdefs depth
	'NNISTD',			# standard deviation for ifdefs
	'USCODE',			# shared code usermade
	'UDERIV',			# derivatives usermade
	'UDESC',			# combination of shared code and derivates by user
	'SCODE',			# shared code "||"
	'DERIV',			# derivative "&&"
	'DESC',				# combination of shared code and derivative
	'HET',				# heterogenous features
	'HOM',				# homogenous features
	'HOHE',				# combination of het and hom features
	'SDEGMEAN',			# shared code degree: mean
	'SDEGSTD',			# shared code degree: standard-deviation
	'TDEGMEAN',			# tangled code degree: mean
	'TDEGSTD',			# tangled code degree: standard-deviation
	# getting basic differentiation of srcml.dtd
	# granularity outer (go)
	'GRANGLO',			# top block extension
	'GRANFU',			# function block extension
	'GRANSTRL',			# struct/union block extension local
	'GRANSTRG',			# struct/union block extension global
	'GRANFUTY',			# sum of GRANFU + GRANST{L|G}
	'GRANBL',			# if/while/for/do block extension
	'GRANEXP',			# condition block extension - includes return
	'GRANST',			# statement extension - includes string concat
	'GRANSIG',			# function parameter extension
	'GRANERR',			# not determined granularity
	# granularity inner (gi)
#	'gmacrogr',			# adding a macro: define, directive
#	'ginablgr',			# adding a named block: function, struct, ...
#	'giunblgr',			# adding a unnamed block: if, while, for, ...
#	'giexpbgr',			# adding a simple expression
#	'gistmbgr',			# adding a simple statement
#	'gipambgr',			# adding a parameter to a function
#	'gierror',			# not categorized extension; count as error
	# operator metrics besides logical operators
	'LOGIC',		# number of logical operators
	'L_AND',		# &&
	'L_OR',			# ||
	'L_NOT',		# !
	'COMP',			# number of comparison operators
	'C_LT',			# <
	'C_LE',			# <=
	'C_GT',			# >
	'C_GE',			# >=
	'C_NE',			# !=
	'C_EQ',			# ==
	'ARITH',		# number of arithmetic operations
	'A_PLUS',		# +
	'A_MINUS',		# -
	'A_MULT',		# *
	'A_DIV',		# /
	'A_MOD',		# %
	'A_SHIFTR',		# >>
	'A_SHIFTL',		# <<
	'BIT',			# number of bit operations
	'B_OR',			# |
	'B_AND',		# &
)
##################################################


##################################################
# options parsing
parser = OptionParser()
parser.add_option("--folder", dest="folder",
		help="input folder [default=.]", default=".")
parser.add_option("--maple", dest="maple", action="store_true",
		default=False, help="make use of maple for checking " \
		"feature expressions equality[default=False]")
parser.add_option("--csp", dest="csp", action="store_true",
		default=False, help="make use of csp solver to check " \
		"feature expression equality [default=False]")
parser.add_option("--str", dest="str", action="store_true",
		default=True, help="make use of simple string comparision " \
		"for checking feature expression equality [default=True]")
parser.add_option("--mapleh", dest="mapleh",
		default="bashir", help="name of the host with maple xmlrpc [default=bashir]")
parser.add_option("--maplep", dest="maplep",
		default="10001", help="port of the host with maple xmlrpc [default=10001]")

(options, args) = parser.parse_args()


##################################################
# helper functions, constants and errors
def _flatten(l):
	"""This function takes a list as input and returns a flatten version
	of the list. So all nested lists are unpacked and moved up to the
	level of the list."""
	i = 0
	while i < len(l):
		while isinstance(l[i], list):
			if not l[i]:
				l.pop(i)
				i -= 1
				break
			else:
				l[i:i+1] = l[i]
		i += 1
	return l


def _collectDefines(d):
	"""This functions adds all defines to a set.
	e.g. #define FEAT_WIN
	also #define FEAT_WIN 12
	but not #define GLIBCVER(x,y,z) ...
	"""
	__defset.add(d[0])
	if __defsetf.has_key(__curfile):
		__defsetf[__curfile].add(d[0])
	else:
		__defsetf[__curfile] = set([d[0]])
	return d


# possible operands:
#   - hexadecimal number
#   - decimal number
#   - identifier
#   - macro function, which is basically expanded via #define
#     to an expression
__string = \
		pypa.Literal('\'').suppress() + \
		pypa.Word(pypa.alphanums+'_\\') + \
		pypa.Literal('\'').suppress()

__hexadec = \
		pypa.Literal('0x').suppress() + \
		pypa.Word(pypa.hexnums).\
		setParseAction(lambda t: str(int(t[0], 16))) + \
		pypa.Optional(pypa.Suppress(pypa.Literal('L')))

__integer = \
		pypa.Optional('~') + \
		pypa.Word(pypa.nums+'-').setParseAction(lambda t: str(int(t[0]))) + \
		pypa.Optional(pypa.Suppress(pypa.Literal('U'))) + \
		pypa.Optional(pypa.Suppress(pypa.Literal('L'))) + \
		pypa.Optional(pypa.Suppress(pypa.Literal('L')))

__identifier = \
		pypa.Word(pypa.alphanums+'_'+'-').setParseAction(_collectDefines)
__arg = pypa.Word(pypa.alphanums+'_')
__args = __arg + pypa.ZeroOrMore(pypa.Literal(',').suppress() + \
		__arg)
__fname = pypa.Word(pypa.alphas, pypa.alphanums + '_')
__function = pypa.Group(__fname + pypa.Literal('(').suppress() + \
		__args + pypa.Literal(')').suppress())


class NoEquivalentSigError(Exception):
	def __init__(self):
		pass
	def __str__(self):
		return ("No equivalent signature found!")

##################################################


def _collapseSubElementsToList(node):
	"""This function collapses all subelements of the given element
	into a list used for getting the signature out of an #ifdef-node."""
	# get all descendants - recursive - children, children of children ...
	itdesc = node.itertext()

	# iterate over the elemtents and add them to a list
	return ''.join([it for it in itdesc])


def _parseFeatureSignatureAndRewriteCSP(sig):
	"""This function parses a given feature-expresson and
	rewrites the expression according to the given __pt mapping.
	This one is used to make use of a csp solver without using
	a predicate."""
	__pt = {
		#'defined' : 'defined_',
		'defined' : '',
		'!' : '!',
		'&&': '&',
		'||': '|',
		'<' : '_lt_',
		'>' : '_gt_',
		'<=': '_le_',
		'>=': '_ge_',
		'==': '_eq_',
		'!=': '_ne_',
		'*' : '_mu_',
		'/' : '_di_',
		'%' : '_mo_',
		'+' : '_pl_',
		'-' : '_mi_',
		'&' : '_ba_',
		'|' : '_bo_',
		'>>': '_sr_',
		'<<': '_sl_',
	}
	mal = list()

	def _rewriteOne(param):
		"""This function returns each one parameter function
		representation for csp."""
		op, ma = param[0]
		mal.append(ma)
		if op == '!': ret = __pt[op] + '(' + ma + ')'
		if op == 'defined': ret = ma
		return  ret

	def _rewriteTwo(param):
		"""This function returns each two parameter function
		representation for csp."""
		mal.extend(param[0][0::2])
		ret = __pt[param[0][1]]
		ret = '(' + ret.join(map(str, param[0][0::2])) + ')'
		return ret

	operand = __string | __hexadec | __integer | \
			__function | __identifier
	compoperator = pypa.oneOf('< > <= >= == !=')
	calcoperator = pypa.oneOf('+ - * / & | << >>')
	expr = pypa.operatorPrecedence(operand, [
		('defined', 1, pypa.opAssoc.RIGHT, _rewriteOne),
		('!',  1, pypa.opAssoc.RIGHT, _rewriteOne),
		(calcoperator, 2, pypa.opAssoc.LEFT, _rewriteTwo),
		(compoperator, 2, pypa.opAssoc.LEFT, _rewriteTwo),
		('&&', 2, pypa.opAssoc.LEFT, _rewriteTwo),
		('||', 2, pypa.opAssoc.LEFT, _rewriteTwo),
	])

	try:
		rsig = expr.parseString(sig)[0]
	except pypa.ParseException, e:
		print('ERROR (parse): cannot parse sig (%s) -- (%s)' %
				(sig, e.col))
		return sig
	except RuntimeError:
		print('ERROR (time): cannot parse sig (%s)' % (sig))
		return sig
	return (mal, ''.join(rsig))


def _parseFeatureSignatureAndRewrite(sig):
	"""This function parses a given feature-signature and rewrites
	the signature according to the given __pt mapping.
	"""
	# this dictionary holds all transformations of operators from
	# the origin (cpp) to the compare (language)
	# e.g. in cpp && stands for the 'and'-operator.
	# the equivalent in maple (which is used for comparison)
	# is '&and'
	# if no equivalence can be found a name rewriting is done
	# e.g. 'defined'
	__pt = {
		#'defined' : 'defined_',
		'defined' : '',
		'!' : '&not',
		'&&': '&and',
		'||': '&or',
		'<' : '<',
		'>' : '>',
		'<=': '<=',
		'>=': '>=',
		'==': '=',
		'!=': '!=',
		'*' : '*',		# needs rewriting with parenthesis
		'/' : '/',
		'%' : '',		# needs rewriting a % b => modp(a, b)
		'+' : '+',
		'-' : '-',
		'&' : '',		# needs rewriting a & b => BitAnd(a, b)
		'|' : '',		# needs rewriting a | b => BitOr(a, b)
		'>>': '>>',		# needs rewriting a >> b => a / (2^b)
		'<<': '<<',		# needs rewriting a << b => a * (2^b)
	}

	def _rewriteOne(param):
		"""This function returns each one parameter function
		representation for maple."""
		if param[0][0] == '!':
			ret = __pt[param[0][0]] + '(' + str(param[0][1]) + ')'
		if param[0][0] == 'defined':
			ret = __pt[param[0][0]] + str(param[0][1])
		return  ret


	def _rewriteTwo(param):
		"""This function returns each two parameter function
		representation for maple."""
		# rewriting rules
		if param[0][1] == '%':
			return 'modp(' + param[0][0] + ',' + param[0][2] + ')'

		ret = ' ' + __pt[param[0][1]] + ' '
		ret = '(' + ret.join(map(str, param[0][0::2])) + ')'

		if param[0][1] in ['<', '>', '<=', '>=', '!=', '==']:
			ret = '(true &and ' + ret + ')'
		return ret

	operand = __string | __hexadec | __integer | \
			__function | __identifier
	compoperator = pypa.oneOf('< > <= >= == !=')
	calcoperator = pypa.oneOf('+ - * / & | << >>')
	expr = pypa.operatorPrecedence(operand, [
		('defined', 1, pypa.opAssoc.RIGHT, _rewriteOne),
		('!',  1, pypa.opAssoc.RIGHT, _rewriteOne),
		(calcoperator, 2, pypa.opAssoc.LEFT, _rewriteTwo),
		(compoperator, 2, pypa.opAssoc.LEFT, _rewriteTwo),
		('&&', 2, pypa.opAssoc.LEFT, _rewriteTwo),
		('||', 2, pypa.opAssoc.LEFT, _rewriteTwo),
	])

	try:
		rsig = expr.parseString(sig)[0]
	except pypa.ParseException, e:
		print('ERROR (parse): cannot parse sig (%s) -- (%s)' %
				(sig, e.col))
		return sig
	except RuntimeError:
		print('ERROR (time): cannot parse sig (%s)' % (sig))
		return sig
	return ''.join(rsig)


def _getMacroSignature(ifdefnode):
	"""This function gets the signature of an ifdef or corresponding macro
	out of the xml-element and its descendants. Since the macros are held
	inside the xml-representation in an own namespace, all descendants
	and their text corresponds to the macro-signature.
	"""
	# get either way the expr-tag for if and elif
	# or the name-tag for ifdef and ifndef,
	# which are both the starting point for signature
	# see the srcml.dtd for more information
	nexpr = []
	res = ''
	ns, tag = __cpprens.match(ifdefnode.tag).groups()

	# get either the expr or the name tag,
	# which is always the second descendant
	if (tag in ['if', 'elif', 'ifdef', 'ifndef']):
		nexpr = [itex for itex in ifdefnode.iterdescendants()]
		if (len(nexpr) == 1):
			res = nexpr[0].tail
		else:
			nexpr = nexpr[1]
			res = ''.join([token for token in nexpr.itertext()])
	return res


def _prologCSV(folder):
	"""prolog of the CSV-output file
	no corresponding _epilogCSV."""
	fd = open(os.path.join(folder, __outputfile), 'w')
	fdcsv = csv.writer(fd, delimiter=';')
	fdcsv.writerow(list(__statsorder._keys))
	return (fd, fdcsv)


def _countNestedIfdefs(root):
	"""This function counts the number of nested ifdefs (conditionals)
	within the source-file."""
	cncur = 0
	cnlist = []
	elements = [it for it in root.iterdescendants()]

	for elem in elements:
		ns, tag = __cpprens.match(elem.tag).groups()
		if ((tag in __conditionals_endif)
				and (ns == __cppnscpp)): cncur -= 1
		if ((tag in __conditionals)
				and (ns == __cppnscpp)):
			cncur += 1
			cnlist.append(cncur)

	if (len(cnlist) > 0):
		nnimax = max(cnlist)
		nnitmp = filter(lambda n: n > 0, cnlist)
		nnimean = pstat.stats.lmean(nnitmp)
	else:
		nnimax = 0
		nnimean = 0
	if (len(cnlist) > 1): nnistd = pstat.stats.lstdev(cnlist)
	else: nnistd = 0
	return (nnimax, nnimean, nnistd)


def _getFeatureSignature(condinhist):
	"""This method returns a feature signature that belongs to the
	current history of conditional inclusions held in condinhist."""
	# we need to rewrite the elements before joining them to one
	# signature; reason is elements like else or elif, which mean
	# basically invert the fname found before
	# rewritelist = [(tag, fname, <invert true|false>)]
	rewritelist = [None]*len(condinhist)
	cur = -1

	for tag, fname in condinhist:
		cur += 1
		if tag == 'if':
			rewritelist[cur] = (tag, fname, False)
		if tag in ['elif', 'else']:
			(t, f, i) = rewritelist[cur-1]
			rewritelist[cur-1] = (t, f, True)
			rewritelist[cur] = (tag, fname, False)

	fsig = ''

	for (tag, fname, invert) in rewritelist:
		if invert:
			fname = '!(' + fname + ')'
		if fsig == '':
			fsig = fname
			continue
		if tag == 'else':
			continue
		if tag in ['if', 'elif']:
			fsig = '(' + fsig + ') && (' + fname + ')'
			continue
	return fsig


def _getASTHistory(node):
	"""This function returns a list with a AST History until
	the given parameter node. The given node is the macro-conditional
	node itself. """
	ancs = [anc for anc in node.iterancestors()]
	asth = []

	for anc in ancs:
		ns, tag = __cpprens.match(anc.tag).groups()
		asth.append(tag)
	return asth


def _getASTFuture(node):
	"""This function returns a list with a AST Future beginning from
	the given parameter node. The given node is the macro-conditional
	node itself."""

	dess = []
	while (node is not None):
		dess += [sib for sib in node.itersiblings(preceding=False)]
		node = node.getparent()

	desh = []
	for des in dess:
		ns, tag = __cpprens.match(des.tag).groups()
		desh.append(tag)
	return desh


def _parseAndAddDefine(node):
	"""This function extracts the identifier and the corresponding
	expansion from define macros. Later on these are used in conditionals
	in order to make them comparable."""
	define = _collapseSubElementsToList(node)

	# match only macro functions, no macro objects
	anytext = pypa.Word(pypa.printables)
	macrodef = pypa.Literal('#define').suppress() + __function + anytext

	try:
		res = macrodef.parseString(define)
	except pypa.ParseException:
		return

	iden = ''.join(map(str, res[0]))
	expn = res[-1]
	para = res[1:-1]
	__macrofuncs[iden] = (para, expn)


def _getFeatures(root):
	"""This function returns all features in the source-file.
	A feature is defined as an enframement of soure-code. The frame
	consists of an ifdef (conditional) and an endif-macro. The function
	returns a tuple with the following format:
	({<feature signature>: (<feature depth>, [<feature code>])},
	 {<feature signature>: [<feature tags-enclosed>]},
	 [(<feature signature>, (<start>, <end>))])

	feature elements: Every feature element reflects one part of a
	feature withing the whole source-code, that is framed by contional
	and endif-macros.
	
	featuresgrinner: All tags from the feature elements (see above).
	featuresgrouter: All tags from the elements arround the feature.
	"""

	def _wrapGrOuterUp(fouter, featuresgrouter, eelem):
		itouter = fouter[-1]				# feature surround tags
		fouter = fouter[:-1]

		selem = itouter[0][1]
		for (sig, elem) in itouter:
			featuresgrouter.append((sig, selem, eelem))
		return (fouter, featuresgrouter)


	def _wrapFeatureUp(features, featuresgrinner, fcode, flist, finner):
		# wrap up the feature
		itsig = flist[-1]				# feature signature
		flist = flist[:-1]
		itcode = fcode[-1]				# feature code
		itcode = itcode.replace('\n\n', '\n')
		itcode = itcode[1:]				# itcode starts with '\n'; del
		fcode = fcode[:-1]
		itinner = finner[-1]				# feature enclosed tags
		finner = finner[:-1]

		# handle the feature code
		if (features.has_key(itsig)):
			features[itsig][1].append(itcode)
		else:
			features[itsig] = (len(flist)+1, [itcode])

		# handle the inner granularity
		featuresgrinner.append((itsig, itinner))
		return (features, featuresgrinner, fcode, flist, finner)

	features = {}			# see above; return value
	featuresgrinner = []	# see above; return value
	featuresgrouter = []	# see above; return value
	flist = []				# holds the features in order
							# list empty -> no features to parse
							# list used as a stack
							# last element = top of stack;
							# and the element we currently
							# collecting source-code lines for
	fouter = []				# holds the xml-nodes of the ifdefs/endifs
							# in order like flist
	fcode = []				# holds the code of the features in
							# order like flist
	finner = []				# holds the tags of the features in
							# order like flist
	condinhist = []			# order of the conditional includes
							# with feature names
	parcon = False			# parse-conditional-flag
	parend = False			# parse-endif-flag
	elelifdepth = 0			# else and elif depth

	# iterate over all tags separately <start>- and <end>-tag
	for event, elem in etree.iterwalk(root, events=("start", "end")):
		ns, tag = __cpprens.match(elem.tag).groups()

		# handling conditionals
		# hitting on conditional-macro
		if ((tag in __conditionals_all)
				and (event == 'start')
				and (ns == __cppnscpp)):	# check the cpp:namespace
			parcon = True

		# hitting next conditional macro; any of ifdef, else or elif
		if ((tag in __conditionals_all)
				and (event == 'end')
				and (ns == __cppnscpp)):	# check the cpp:namespace
			parcon = False

			# with else or elif we finish up the last if, therefor
			# we can applicate the wrapup
			if ((tag in __conditionals_else)
					or (tag in __conditionals_elif)):
				(features, featuresgrinner,
						fcode, flist, finner) = _wrapFeatureUp(features,
						featuresgrinner, fcode, flist, finner)

			fname = _getMacroSignature(elem)
			if fname: condinhist.append((tag, fname))
			else: condinhist.append((tag, ''))

			fsig = _getFeatureSignature(condinhist)
			if (tag in __conditionals): fouter.append([])
			fouter[-1] += ([(fsig, elem)])
			flist.append(fsig)
			fcode.append('')
			finner.append([])

		# hitting end-tag of elif-macro
		if ((tag in __conditionals_elif)
				and (event == 'end')
				and (ns == __cppnscpp)):
			parcon = False

		# hitting end-tag of define-macro
		if ((tag in __macro_define) \
				and (event == 'end') \
				and (ns == __cppnscpp)):
			_parseAndAddDefine(elem)

		# iterateting in subtree of conditional-node
		if parcon:
			continue

		# handling endif-macro
		# hitting an endif-macro start-tag
		if ((tag in __conditionals_endif) \
				and (event == "start") \
				and (ns == __cppnscpp)):	# check the cpp:namespace
			parend = True

		# hitting the endif-macro end-tag
		if ((tag in __conditionals_endif) \
				and (event == "end") \
				and (ns == __cppnscpp)):	# check the cpp:namespace
			parend = False

			(features, featuresgrinner, fcode, flist, finner) = \
				_wrapFeatureUp(features, featuresgrinner,
				fcode, flist, finner)
			(fouter, featuresgrouter) = _wrapGrOuterUp(fouter,
			featuresgrouter, elem)

			while (condinhist[-1][0] != 'if'):
				condinhist = condinhist[:-1]
			condinhist = condinhist[:-1]

		# iterating the endif-node subtree
		if parend:
			continue

		# collect the source-code of the feature
		if (len(flist)):
			if ((event == "start") and (elem.text)):
				fcode[-1] += elem.text
			if ((event == "end") and (elem.tail)):
				fcode[-1] += elem.tail

			if (ns == __cppnsdef or tag not in __conditionals_all):
				finner[-1].append((tag, event, elem.sourceline))

	return (features, featuresgrinner, featuresgrouter)


def _getOuterGranularity(fnodes):
	"""This function determines and returns the outer granularity
	metrics for each feature. Therefore we get a list holding all
	features in order and their start and end node (conditionals)
	from the xml-tree."""
	grouter = list()

	for (sig, selem, eelem) in fnodes:
		tags = _getASTHistory(selem)[:-1]	# cut of unit-tag
		grouter.append((sig, tags, selem.sourceline))
	return grouter


def _getOuterGranularityStats(lgran):
	"""This function determines the granularity level of the
	given lgran elements. We distinguish the following levels:
	- outer block gran (function, struct, union)
	- inner block gran (if, for, while, do)
	- expression gran (condition in if, for, while, do)
	- statement gran
	- parameter gran
	"""
	gotopbgr = 0
	gofunbgr = 0
	gostrbrl = 0		# local
	gostrbrg = 0		# global
	goinnbgr = 0
	goexpbgr = 0
	gostmbgr = 0
	gopambgr = 0
	goerror  = 0

	for (sig, gran, line) in lgran:
		if len(gran) == 0:
			gotopbgr += 1
			continue
		if gran[0] in ['block']:
			if len(gran) == 1:	# configure the method signature
				gofunbgr += 1
				continue
			if gran[1] in ['function', 'extern', 'block']:
				gofunbgr += 1
			elif gran[1] in ['struct', 'union',
					'enum']:	# test_struct_union_enum.c
				if 'function' in gran[2:]: gostrbrl += 1
				else: gostrbrg += 1
			elif gran[1] in ['expr']:
				if 'function' in gran[2:]: gostrbrl += 1
				else: gostrbrg += 1
			elif gran[1] in ['while', 'for', 'then', 'do',
					'else', 'switch', 'case', 'default']:
				goinnbgr += 1				# test_loop.c
			elif gran[1] in ['decl']:		# test_struct_union_enum.c
				if 'function' in gran[3:]: gostrbrl += 1
				else: gostrbrg += 1
			else:
				print('ERROR: gran (%s) at this '
						'level unknown (line %s)' % (gran, line))
				goerror += 1
			continue
		elif gran[0] in ['expr']:
			if gran[1] in ['expr_stmt']:				# test_stmt.c
				gostmbgr += 1
			elif gran[1] in ['condition', 'return']:	# test_condition.c
				goexpbgr += 1
			elif gran[1] in ['argument']:				# test_call.c
				gostmbgr += 1
			elif gran[1] in ['block']:
				if 'function' in gran[2:]: gostrbrl += 1
				else: gostrbrg += 1
			elif gran[1] in ['init', 'index']:			# test_stmt.c
				gostmbgr += 1
			else:
				print('ERROR: gran (%s) at this level'
						'unknown (line %s)' % (gran, line))
				goerror += 1
			continue
		elif gran[0] in ['while', 'do']:				# test_loop.c
			goinnbgr += 1
			continue
		elif gran[0] in ['expr_stmt'] and len(gran) == 1:
			gostmbgr += 1
			continue
		elif gran[:3] == ['expr_stmt', 'block', 'struct']:
			if 'function' in gran[2:]: gostrbrl += 1
			else: gostrbrg += 1
			continue
		elif gran[0] in ['decl_stmt']:			# test_stmt.c
			gostmbgr += 1
			continue
		elif gran[0] in ['condition']:			# test_condition.c
			goexpbgr += 1
			continue
		elif gran[0] in ['if', 'else', 'case', 'default',
				'then', 'for']:	# test_condition.c
			goinnbgr += 1
			continue
		elif gran[0] in ['parameter_list',
				'argument_list']:		# test_call.c
			gopambgr += 1
			continue
		elif gran[0] in ['argument'] and gran[1] in ['argument_list']:
			gostmbgr += 1
		elif gran[0] in ['init'] and gran[1] in ['decl']:	# test_stmt.c
			gostmbgr += 1
			continue
		elif gran[0] in ['function']:			# function prototype
			continue
		else:
			print('ERROR: outer granularity (%s, %s) not recognize!' % \
					(gran, line))
			goerror += 1

	return (gotopbgr, gofunbgr, gostrbrl, gostrbrg,
			goinnbgr, goexpbgr, gostmbgr, gopambgr, goerror)


def _getInnerGranularityStats(igran):
	"""This method returns a tuple with the information about the
	inner granularity. We distinguish the following granularities:
	- adding a named block, ... (a whole block with a name:
			function, type, ...)
	- adding a unnamed block, ... (a block without a name:
			if, while, for, ...)
	- adding a simple statement
	- adding a expression
	- adding a parameter
	"""
	gmacrogr = 0
	ginablgr = 0
	giunblgr = 0
	giexpbgr = 0
	gistmbgr = 0
	gipambgr = 0
	gierror = 0

	skiptilltag = ''

	for (sig, gran) in igran:
		for (tag, event, line) in gran:
			if (skiptilltag != ''):
				if (tag == skiptilltag[0]
						and event == 'end'
						and line == skiptilltag[2]):
					skiptilltag = ''
				continue

			if tag in ['name', 'endif']: continue
			elif tag in ['define', 'directive', 'include',
					'macro', 'undef']:
				gmacrogr += 1
			elif tag in ['struct', 'union', 'enum', 'function', 'extern',
					'function_decl', 'decl_stmt', 'typedef']:
				ginablgr += 1
			elif tag in ['if', 'while', 'return', 'then',
					'for', 'do', 'case', 'else', 'block']:
				giunblgr += 1
			elif tag in ['param', 'argument']:
				gipambgr += 1
			elif tag in ['expr']:
				giexpbgr += 1
			elif tag in ['condition']:
				giexpbgr += 1
			elif tag in ['expr_stmt', 'decl', 'init']:
				gistmbgr += 1
			else:
				print('ERROR: inner granularity (%s, %s, %s)) '
						'not recognized!' % (tag, event, line))
				gierror += 1
				continue
			if event == 'start': skiptilltag = (tag, event, line)

	return (gmacrogr, ginablgr, giunblgr, giexpbgr, gistmbgr, gierror)


def _getFeatureStats(features):
	"""This function determines and returns the following statistics
	about the features:
		- nof			# number of features
		- nod			# number of defines
		- lof			# lines of feature code (sum)
		- lofmin		# minimum line of feature code
		- lofmax		# maximum number of feature code lines
		- lofmean		# mean of feature code lines
		- lofstd		# std-deviation of feature code lines
	"""
	nof = 0
	lof = 0
	nod = 0
	lofmin = -1			# feature-code can be empty
	lofmax = 0
	floflist = []		# feature-code; needed for lofmean, lofstd
	lofmean = 0
	lofstd = 0
	nof = len(features.keys())
	tmp = [item for (depth, item) in features.itervalues()]
	tmpl = list(tmp)
	tmp = _flatten(tmp)
	floflist = map(lambda n: n.count('\n'), tmp)

	if (len(floflist)):
		lofmin = min(floflist)
		lofmax = max(floflist)
		lof = reduce(lambda m,n: m+n, floflist)
		lofmean = pstat.stats.lmean(floflist)

	if (len(floflist) > 1):
		lofstd = pstat.stats.lstdev(floflist)

	return (nof, nod, lof, lofmin, lofmax, lofmean, lofstd)


def _getFeaturesDepthOne(features):
	"""This function returns all features that have the depth of one."""
	nof1 = dict()
	nof1 = filter(lambda (sig, (depth, code)): depth == 1, features.iteritems())
	return nof1


def _distinguishFeatures(features):
	"""This function returns a tuple with dicts, each holding
	one type of feature. The determination is according to the
	given macro-signatures. Following differentiation according
	to the macro-signatures:
	1. "||" -> shared code
	2. "&&" -> derivative
	3. "||" & "&&" -> ??

	Further more a differentiation according to the feature-code
	is also done. We differ here:
	1. het -> all code feature code excerpts are different
	2. hom -> all feature code are the same
	3. hethome -> the feature code is a mix of both
	"""

	def _compareFeatureCode(fcode):
		"""This function compares the each part, of the code the
		belongs to a feature with all the other parts. We do this
		in order to find out, what kind of "introduction" is made
		at the feature signature points. For instance:
		"""
		fcodes = set(fcode)

		if (len(fcodes) == 1 and len(fcode) > 1):
			return "hom"

		if (len(fcode) == len(fcodes)):
			return "het"

		if (len(fcode) > len(fcodes)):
			return "hethom"

	scode = {}
	deriv = {}
	desc = {}
	het = {}
	hom = {}
	hethom = {}

	for (key, (depth, item)) in features.iteritems():
		# distinguish according to feature-signature
		# shared code
		if ('||' in key and (not '&&' in key)):
			scode[key] = item

		# derivative only &&
		if ('&&' in key and (not '||' in key)):
			deriv[key] = item

		# combination shared code and derivative
		if ('&&' in key and '||' in key):
			desc[key] = item

		# distinguish according to feature-code
		ret = _compareFeatureCode(item)
		if (ret == "het"):
			het[key] = item
		if (ret == "hom"):
			hom[key] = item
		if (ret == "hethom"):
			hethom[key] = item

	return (scode, deriv, desc, het, hom, hethom)


def _getUserdefinedDerivatives(root):
	"""This method takes the xml-root and iteratates over the elements
	in order to find all user defined derivates. These derivatives are
	the result of nested ifdefs."""
	uscode = 0
	uderiv = 0
	udesc  = 0

	for (action, elem) in etree.iterparse(root):
		ns, tag = __cpprens.match(elem.tag).groups()
		if (ns == __cppnscpp):
			sig = _getMacroSignature(elem)

			if ('||' in sig and (not '&&' in sig)):
				uscode += 1
			if ('&&' in sig and (not '||' in sig)):
				uderiv += 1
			if ('&&' in sig and '||' in sig):
				udesc += 1
	return (uscode, uderiv, udesc)


def _getNumOfOperators(features):
	"""This method returns a tuple with the total occurance of certain
	operators (compare, arithmetic and bit-operations)."""
	logic = 0			# all logic operators
	l_and = 0			# &&
	l_or = 0			# ||
	l_not = 0			# !
	comp = 0			# all comparison
	c_lt = 0			# <
	c_le = 0			# <=
	c_gt = 0			# >
	c_ge = 0			# >=
	c_ne = 0			# !=
	c_eq = 0			# ==
	arith = 0			# all arithmetic
	a_plus = 0			# +
	a_minus = 0			# -
	a_mult = 0			# *
	a_div = 0			# /
	a_mod = 0			# %
	a_shiftr = 0		# >>
	a_shiftl = 0		# <<
	bit = 0				# all bit-operations
	b_or = 0			# |
	b_and = 0			# &

	re_c_lt = re.compile('[^<]<[^<=]')
	re_c_gt = re.compile('[^>]>[^>=]')
	re_b_or = re.compile('[^\|]\|[^\|]')
	re_b_and = re.compile('[^&]&[^&]')

	for sig in features.keys():
		if '&&' in sig: l_and += sig.count('&&')
		if '||' in sig: l_or += sig.count('||')
		if '!' in sig: l_not += sig.count('!')
		if '<<' in sig: a_shiftl += sig.count('<<')
		if '<=' in sig: c_le += sig.count('<=')
		if '>>' in sig: a_shiftr += sig.count('>>')
		if '>=' in sig: c_ge += sig.count('>=')
		if '!=' in sig: c_ne += sig.count('!=')
		if '==' in sig: c_eq += sig.count('==')
		if '+' in sig: a_plus += sig.count('+')
		if '-' in sig: a_minus += sig.count('-')
		if '*' in sig: a_mult += sig.count('*')
		if '/' in sig: a_div += sig.count('/')
		if '%' in sig: a_mod += sig.count('%')
		if re_c_lt.findall(sig):
			c_lt += len(re_c_lt.findall(sig))	# single '<'
		if re_c_gt.findall(sig):
			c_gt += len(re_c_gt.findall(sig))	# single '>'
		if re_b_or.findall(sig):
			b_or += len(re_b_or.findall(sig))	# single '|'
		if re_b_and.findall(sig):
			b_and += len(re_b_and.findall(sig))	# single '&'

	logic = l_and + l_or + l_not
	comp = c_lt + c_le + c_gt + c_ge + c_eq
	arith = a_plus + a_minus + a_mult + a_div + \
			a_mod + a_shiftr + a_shiftl
	bit = b_or + b_and

	return (logic, l_and, l_or, l_not, comp, c_lt,
			c_le, c_gt, c_ge, c_eq, c_ne,  arith,
			a_plus, a_minus, a_mult, a_div, a_mod,
			a_shiftr, a_shiftl, bit, b_or, b_and)


def _getNumOfDefines(defset):
	"""This method returns the number of defines, that have the following
	structure:
	#define FEAT_A
	#define FEAT_B 5
	Both defines are macro-objects. macro-functions like the following
	are not considered.
	#define CHECKVERSION(x,y,z) x*100+y*10+z
	All determined elements are derived from the ifdef macros.
	"""
	# basic operation of this function is to check __defset against
	# __macrofuncs
	funcmacros = __macrofuncs.keys()
	funcmacros = map(lambda n: n.split('(')[0], funcmacros)
	funcmacros = set(funcmacros)

	return len((defset - funcmacros))


def _getScatteringTanglingDegrees(sigs, defines):
	"""This method returns the mean and the standard-deviation of
	defines according to the given mapping of a define to occurances
	in the signatures. The input is all feature-signatures and
	a all defines.

	The measurement of scattering and tangling degree is error-prone
	since a define that is used in conditional inclusions might not
	be defined at the moment of usage.
	"""

	def __add(x, y):
		"""This method is a helper function to add values
		of lists pairwise. See below for more information."""
		return x+y

	scat = list()			# relation define to signatures
	tang = [0]*len(sigs)	# signatures overall
	for d in defines:
		dre = re.compile(r'\b'+d+r'\b')		# using word boundaries
		vec = map(lambda s: not dre.search(s) is None, sigs)
		scat.append(vec.count(True))
		tang = map(__add, tang, vec)

	if (len(scat)): sdegmean = pstat.stats.lmean(scat)
	else: sdegmean = 0
	if (len(scat) > 1): sdegstd = pstat.stats.lstdev(scat)
	else: sdegstd = 0

	if (len(tang)): tdegmean = pstat.stats.lmean(tang)
	else: tdegmean = 0
	if (len(tang) > 1): tdegstd = pstat.stats.lstdev(tang)
	else: tdegstd = 0

	return (sdegmean, sdegstd, tdegmean, tdegstd)


def _getGranularityStats(fcodetags):
	"""This method returns a tuple of NOO with decl_stmt, expr_stmt
	and signature changes.
	TODO which granularity to use?
	if   -> block
	for  -> block
	expr -> expression
	"""
	_interestingtags = [
		'define',			# define statement
		'include',			# include statement
		'decl_stmt',		# declaration statement
		'expr_stmt',		# expression statement
		'function_decl',	# function declaration
		'parameter_list',	# parameter list
		'param',			# parameter
	]
	_curskiptag = None		# holds the element we are going to skip for
	_skipedtags = []		# if we find one of the elements above, we
							# start skipping the following tags, until
							# the corresponding endtag is found; if we
							# do not find the endtag, we are going to
							# start analyzing the elements in _skipedtags
	granstats = dict()

	for tag in _interestingtags:
		granstats[tag] = 0

	for (fsig, ftagsl) in fcodetags:
		_curskiptag = None
		_skipedtags = []
		for (tag, event, sourceline) in ftagsl:
			if _curskiptag == None and tag in _interestingtags:
				_curskiptag = tag
				continue
			if _curskiptag != None and tag == _curskiptag:
				granstats[tag] = granstats[tag] + 1
				_curskiptag = None
				_skipedtags = []
				continue
			if _curskiptag != None:
				_skipedtags.append((tag, sourceline))
				continue
#		if _skipedtags != []:
#			print("ERROR: ", _skipedtags)
	return granstats


def _checkForEquivalentSig(l, sig):
	"""This method takes a list of signatures and checks sig for an
	equivalent signature. If no equivalent signature is found this
	method raises an error."""
	if options.maple:
		mserv = xmlrpclib.ServerProxy(':'.join(['http://' + options.mapleh, options.maplep]))

	def _checkSigEquivalence(sig1, sig2):
		"""This function checks the equivalence of two signatures.
		It uses an xmlrpc-call on troi.fim.uni-passau.de."""
		global __errorfexp
		global __errormatch
		if not (sig1 and sig2):			# ommit empty signatures
			return False

		if options.maple:
			ret = mserv.checkEquivalence(sig1, sig2)
			if ret == 'error':
				return False
				__errormatch.append((sig1, sig2))
			if ret == 'true':
				return True
			if ret == 'false':
				return False
			if ret == None:
				return sig1 == sig2
			if (sig1 == sig2) != ret:
				print("INFO: maple-signature match is different")
				__errorfexp += 1
		if options.str:
			return sig1 == sig2

	# _checkForEquivalentSig
	for it in l:
		if _checkSigEquivalence(it, sig):
			return it
	raise NoEquivalentSigError()


def apply(folder):
	"""This function applies the analysis to all xml-files in that
	directory and take the results and joins them together. Results
	are getting written into the fdcsv-file."""
	# overall status variables
	sigmap = {}				# {<converted sig>: [<equivalent sigs>]}
	afeatures = {}			# identified features; {<sig>: (depth, [code])}

	def _mergeFeatures(ffeatures):
		"""This function merges the, with the parameter given
		dictionary (ffeatures) to the afeatures (overall-features)."""
		for (sig, (depth, code)) in ffeatures.iteritems():
			psig = _parseFeatureSignatureAndRewrite(sig)

			try:
				sigmatch = _checkForEquivalentSig(sigmap.keys(), psig)
				(tmpdepth, tmpcode) = afeatures[sigmap[sigmatch][0]]
#				if (tmpdepth != depth):
#					print("INFO: depths of feature fragments do not" +
#							" match (%s, %s)!" % (str(tmpdepth), str(depth)))
				tmpdepth = min(tmpdepth, depth)
				tmpcode += code
				afeatures[sigmap[sigmatch][0]] = (tmpdepth, tmpcode)
				sigmap[sigmatch].append(sig)
			except NoEquivalentSigError:
				# mergedfeatures get the depth of minus one
				# so this way need to make less amount of changes here
				afeatures[sig] = (depth, list(code))
				sigmap[psig] = [sig]

	# outputfile
	fd, fdcsv = _prologCSV(folder)
	fdfeat = open(os.path.join(folder, __outputfexp), 'w')

	global __curfile
	fcount = 0
	files = os.listdir(folder)
	files = filter(lambda n: os.path.splitext(n)[1] == '.xml', files)
	fstats = [None]*len(__statsorder._keys)
	ftotal = len(files)

	# get statistics for all files; write results into csv
	# and merge the features
	for file in files:
		__curfile = file
		fcount += 1
		try:
			tree = etree.parse(file)
		except etree.XMLSyntaxError:
			print("ERROR: cannot parse (%s). Skipping this file." % file)
			continue

		print('INFO: parsing file (%5d) of (%5d) -- (%s).' % (fcount, ftotal, file))

		root = tree.getroot()
		(features, featuresgrinner, featuresgrouter) = _getFeatures(root)
		_mergeFeatures(features)

		# granularity stats
		grouter = _getOuterGranularity(featuresgrouter)
		(gotopbgr, gofunbgr, gostrbrl, gostrbrg,
		goinnbgr, goexpbgr, gostmbgr, gopambgr, goerror) = \
				_getOuterGranularityStats(grouter)
		_getInnerGranularityStats(featuresgrinner)
		fstats[__statsorder.GRANGLO.index] = gotopbgr
		fstats[__statsorder.GRANFU.index] = gofunbgr
		fstats[__statsorder.GRANSTRL.index] = gostrbrl
		fstats[__statsorder.GRANSTRG.index] = gostrbrg
		fstats[__statsorder.GRANFUTY.index] = gofunbgr+gostrbrl+gostrbrg
		fstats[__statsorder.GRANBL.index] = goinnbgr
		fstats[__statsorder.GRANEXP.index] = goexpbgr
		fstats[__statsorder.GRANST.index] = gostmbgr
		fstats[__statsorder.GRANSIG.index] = gopambgr
		fstats[__statsorder.GRANERR.index] = goerror
		grinner = _getGranularityStats(featuresgrinner)

		# general stats
		fstats[__statsorder.FILENAME.index] = file
		(nnimax, nnimean, nnistd) = _countNestedIfdefs(root)
		fstats[__statsorder.NNIMAX.index] = nnimax
		fstats[__statsorder.NNIMEAN.index] = nnimean
		fstats[__statsorder.NNISTD.index] = nnistd
		tmp = [it for it in root.iterdescendants()]

		if (len(tmp)): floc = tmp[-1].sourceline
		else: floc = 0

		fstats[__statsorder.LOC.index] = floc

		# feature-amount
		(nof, nod, lof, lofmin, lofmax, lofmean, lofstd) = \
				_getFeatureStats(features)
		fstats[__statsorder.NOFE.index] = nof
		if __defsetf.has_key(__curfile):
			fstats[__statsorder.NOFC.index] = \
					_getNumOfDefines(__defsetf[__curfile])
		else:
			fstats[__statsorder.NOFC.index] = 0
		fstats[__statsorder.LOF.index] = lof
		nof1 = _getFeaturesDepthOne(features)
		fstats[__statsorder.NOFE1.index] = len(nof1)

		if lofmin == -1: lofmin = 0

		fstats[__statsorder.LOFMIN.index] = lofmin
		fstats[__statsorder.LOFMAX.index] = lofmax
		fstats[__statsorder.LOFMEAN.index] = lofmean
		fstats[__statsorder.LOFSTD.index] = lofstd

		# user defined derivatives and shared code
		(uscode, uderiv, udesc) = _getUserdefinedDerivatives(file)
		fstats[__statsorder.USCODE.index] = uscode
		fstats[__statsorder.UDERIV.index] = uderiv
		fstats[__statsorder.UDESC.index] = udesc

		# feature differentiation
		(scode, deriv, desc, het, hom, hethom) = \
				_distinguishFeatures(features)
		fstats[__statsorder.SCODE.index] = len(scode.keys())
		fstats[__statsorder.DERIV.index] = len(deriv.keys())
		fstats[__statsorder.DESC.index] = len(desc.keys())
		fstats[__statsorder.HET.index] = len(het.keys())
		fstats[__statsorder.HOM.index] = len(hom.keys())
		fstats[__statsorder.HOHE.index] = len(hethom.keys())

		# scattering and tangling
		# not useful to compute the scattering per file, since a feature names
		# may be defined later
		#if __defsetf.has_key(__curfile):
		#	(sdegmean, sdegstd, tdegmean, tdegstd) = \
		#			_getScatteringTanglingDegrees(features.keys(),
		#			list(__defsetf[__curfile]))
		#else: (sdegmean, sdegstd, tdegmean, tdegstd) = (0, 0, 0, 0)
		(sdegmean, sdegstd, tdegmean, tdegstd) = (0, 0, 0, 0)
		fstats[__statsorder.SDEGMEAN.index] = sdegmean
		fstats[__statsorder.SDEGSTD.index] = sdegstd
		fstats[__statsorder.TDEGMEAN.index] = tdegmean
		fstats[__statsorder.TDEGSTD.index] = tdegstd

		# feature parametrization, compare-operator, arithmetic-operator
		(logic, l_and, l_or, l_not, comp, c_lt, c_le, c_gt, c_ge,
				c_eq, c_ne, arith, a_plus, a_minus, a_mult, a_div,
				a_mod, a_shiftr, a_shiftl, bit, b_or, b_and) = \
				_getNumOfOperators(features)
		fstats[__statsorder.LOGIC.index] = logic
		fstats[__statsorder.L_AND.index] = l_and
		fstats[__statsorder.L_OR.index] = l_or
		fstats[__statsorder.L_NOT.index] = l_not
		fstats[__statsorder.COMP.index] = comp
		fstats[__statsorder.C_LT.index] = c_lt
		fstats[__statsorder.C_LE.index] = c_le
		fstats[__statsorder.C_GT.index] = c_gt
		fstats[__statsorder.C_GE.index] = c_ge
		fstats[__statsorder.C_NE.index] = c_ne
		fstats[__statsorder.C_EQ.index] = c_eq
		fstats[__statsorder.ARITH.index] = arith
		fstats[__statsorder.A_PLUS.index] = a_plus
		fstats[__statsorder.A_MINUS.index] = a_minus
		fstats[__statsorder.A_MULT.index] = a_mult
		fstats[__statsorder.A_DIV.index] = a_div
		fstats[__statsorder.A_MOD.index] = a_mod
		fstats[__statsorder.A_SHIFTR.index] = a_shiftr
		fstats[__statsorder.A_SHIFTL.index] = a_shiftl
		fstats[__statsorder.BIT.index] = bit
		fstats[__statsorder.B_OR.index] = b_or
		fstats[__statsorder.B_AND.index] = b_and

		fdcsv.writerow(fstats)

	# writing convinience functions
	fnum = len(files)+1			# +1 for the header of the table
	excelcols = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J',
			'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U',
			'V', 'W', 'X', 'Y', 'Z', 'AA', 'AB', 'AC', 'AD', 'AE',
			'AF', 'AG', 'AH', 'AI', 'AJ', 'AK', 'AL', 'AM', 'AN',
			'AO', 'AP', 'AQ', 'AR', 'AS', 'AT', 'AU', 'AV', 'AW',
			'AX', 'AY', 'AZ', 'BA', 'BB', 'BC', 'BD', 'BE', 'BF',
			'BG', 'BH', 'BI', 'BJ', 'BK', 'BL', 'BM', 'BN', 'BO',
			'BP', 'BQ', 'BR', 'BS', 'BT', 'BU', 'BV', 'BW', 'BX',
			'BY', 'BZ']
	excelfunc = [None]*len(__statsorder._keys)
	excelfunc[__statsorder.FILENAME.index] = "FUNCTIONS"
	excelfunc[__statsorder.LOC.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.LOC.index], \
			excelcols[__statsorder.LOC.index], fnum)
	excelfunc[__statsorder.NOFE.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.NOFE.index], \
			excelcols[__statsorder.NOFE.index], fnum)
	excelfunc[__statsorder.NOFE1.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.NOFE1.index], \
			excelcols[__statsorder.NOFE1.index], fnum)
	excelfunc[__statsorder.LOF.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.LOF.index], \
			excelcols[__statsorder.LOF.index], fnum)
	excelfunc[__statsorder.LOFMIN.index] = "=MIN(%s2:%s%s)" % \
			(excelcols[__statsorder.LOFMIN.index], \
			excelcols[__statsorder.LOFMIN.index], fnum)
	excelfunc[__statsorder.LOFMAX.index] = "=MAX(%s2:%s%s)" % \
			(excelcols[__statsorder.LOFMAX.index], \
			excelcols[__statsorder.LOFMAX.index], fnum)
	excelfunc[__statsorder.NNIMAX.index] = "=MAX(%s2:%s%s)" % \
			(excelcols[__statsorder.NNIMAX.index], \
			excelcols[__statsorder.NNIMAX.index], fnum)
	excelfunc[__statsorder.NNIMEAN.index] = "=SUM(%s2:%s%s)/(SUMPRODUCT((%s2:%s%s<>0)*1))" % \
			(excelcols[__statsorder.NNIMEAN.index], \
			excelcols[__statsorder.NNIMEAN.index], fnum, \
			excelcols[__statsorder.NNIMEAN.index], \
			excelcols[__statsorder.NNIMEAN.index], fnum)
	excelfunc[__statsorder.NNISTD.index] = "=SUM(%s2:%s%s)/(SUMPRODUCT((%s2:%s%s<>0)*1))" % \
			(excelcols[__statsorder.NNISTD.index], \
			excelcols[__statsorder.NNISTD.index], fnum, \
			excelcols[__statsorder.NNIMEAN.index],			# it might be that mean 1 std 0
			excelcols[__statsorder.NNIMEAN.index], fnum)	# therefore we use mean here
	excelfunc[__statsorder.USCODE.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.USCODE.index], \
			excelcols[__statsorder.USCODE.index], fnum)
	excelfunc[__statsorder.UDERIV.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.UDERIV.index], \
			excelcols[__statsorder.UDERIV.index], fnum)
	excelfunc[__statsorder.UDESC.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.UDESC.index], \
			excelcols[__statsorder.UDESC.index], fnum)
	excelfunc[__statsorder.SCODE.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.SCODE.index], \
			excelcols[__statsorder.SCODE.index], fnum)
	excelfunc[__statsorder.DERIV.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.DERIV.index], \
			excelcols[__statsorder.DERIV.index], fnum)
	excelfunc[__statsorder.DESC.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.DESC.index], \
			excelcols[__statsorder.DESC.index], fnum)
	excelfunc[__statsorder.HET.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.HET.index], \
			excelcols[__statsorder.HET.index], fnum)
	excelfunc[__statsorder.HOM.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.HOM.index], \
			excelcols[__statsorder.HOM.index], fnum)
	excelfunc[__statsorder.HOHE.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.HOHE.index], \
			excelcols[__statsorder.HOHE.index], fnum)
	excelfunc[__statsorder.LOGIC.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.LOGIC.index], \
			excelcols[__statsorder.LOGIC.index], fnum)
	excelfunc[__statsorder.L_AND.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.L_AND.index], \
			excelcols[__statsorder.L_AND.index], fnum)
	excelfunc[__statsorder.L_OR.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.L_OR.index], \
			excelcols[__statsorder.L_OR.index], fnum)
	excelfunc[__statsorder.L_NOT.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.L_NOT.index], \
			excelcols[__statsorder.L_NOT.index], fnum)
	excelfunc[__statsorder.COMP.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.COMP.index], \
			excelcols[__statsorder.COMP.index], fnum)
	excelfunc[__statsorder.C_LT.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.C_LT.index], \
			excelcols[__statsorder.C_LT.index], fnum)
	excelfunc[__statsorder.C_LE.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.C_LE.index], \
			excelcols[__statsorder.C_LE.index], fnum)
	excelfunc[__statsorder.C_GT.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.C_GT.index], \
			excelcols[__statsorder.C_GT.index], fnum)
	excelfunc[__statsorder.C_GE.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.C_GE.index], \
			excelcols[__statsorder.C_GE.index], fnum)
	excelfunc[__statsorder.C_NE.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.C_NE.index], \
			excelcols[__statsorder.C_NE.index], fnum)
	excelfunc[__statsorder.C_EQ.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.C_EQ.index], \
			excelcols[__statsorder.C_EQ.index], fnum)
	excelfunc[__statsorder.ARITH.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.ARITH.index], \
			excelcols[__statsorder.ARITH.index], fnum)
	excelfunc[__statsorder.A_PLUS.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.A_PLUS.index], \
			excelcols[__statsorder.A_PLUS.index], fnum)
	excelfunc[__statsorder.A_MINUS.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.A_MINUS.index], \
			excelcols[__statsorder.A_MINUS.index], fnum)
	excelfunc[__statsorder.A_MULT.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.A_MULT.index], \
			excelcols[__statsorder.A_MULT.index], fnum)
	excelfunc[__statsorder.A_DIV.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.A_DIV.index], \
			excelcols[__statsorder.A_DIV.index], fnum)
	excelfunc[__statsorder.A_MOD.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.A_MOD.index], \
			excelcols[__statsorder.A_MOD.index], fnum)
	excelfunc[__statsorder.A_SHIFTR.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.A_SHIFTR.index], \
			excelcols[__statsorder.A_SHIFTR.index], fnum)
	excelfunc[__statsorder.A_SHIFTL.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.A_SHIFTL.index], \
			excelcols[__statsorder.A_SHIFTL.index], fnum)
	excelfunc[__statsorder.BIT.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.BIT.index], \
			excelcols[__statsorder.BIT.index], fnum)
	excelfunc[__statsorder.B_AND.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.B_AND.index], \
			excelcols[__statsorder.B_AND.index], fnum)
	excelfunc[__statsorder.B_OR.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.B_OR.index], \
			excelcols[__statsorder.B_OR.index], fnum)
	excelfunc[__statsorder.GRANGLO.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.GRANGLO.index], \
			excelcols[__statsorder.GRANGLO.index], fnum)
	excelfunc[__statsorder.GRANFU.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.GRANFU.index], \
			excelcols[__statsorder.GRANFU.index], fnum)
	excelfunc[__statsorder.GRANSTRL.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.GRANSTRL.index], \
			excelcols[__statsorder.GRANSTRL.index], fnum)
	excelfunc[__statsorder.GRANSTRG.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.GRANSTRG.index], \
			excelcols[__statsorder.GRANSTRG.index], fnum)
	excelfunc[__statsorder.GRANFUTY.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.GRANFUTY.index], \
			excelcols[__statsorder.GRANFUTY.index], fnum)
	excelfunc[__statsorder.GRANBL.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.GRANBL.index], \
			excelcols[__statsorder.GRANBL.index], fnum)
	excelfunc[__statsorder.GRANEXP.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.GRANEXP.index], \
			excelcols[__statsorder.GRANEXP.index], fnum)
	excelfunc[__statsorder.GRANST.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.GRANST.index], \
			excelcols[__statsorder.GRANST.index], fnum)
	excelfunc[__statsorder.GRANSIG.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.GRANSIG.index], \
			excelcols[__statsorder.GRANSIG.index], fnum)
	excelfunc[__statsorder.GRANERR.index] = "=SUM(%s2:%s%s)" % \
			(excelcols[__statsorder.GRANERR.index], \
			excelcols[__statsorder.GRANERR.index], fnum)
	fdcsv.writerow(excelfunc)

	# overall - stats
	astats = [None]*len(__statsorder._keys)
	(nofe, nofc, lof, lofmin, lofmax, lofmean, lofstd) = \
			_getFeatureStats(afeatures)
	(sdegmean, sdegstd, tdegmean, tdegstd) = \
			_getScatteringTanglingDegrees(_flatten(sigmap.values()),
			list(__defset))
	nofe1 = _getFeaturesDepthOne(afeatures)

	astats[__statsorder.FILENAME.index] = "ALL - MERGED"
	astats[__statsorder.NOFE.index] = nofe
	astats[__statsorder.NOFE1.index] = len(nofe1)
	astats[__statsorder.NOFC.index] = _getNumOfDefines(__defset)
	astats[__statsorder.LOF.index] = lof
	astats[__statsorder.LOFMIN.index] = lofmin
	astats[__statsorder.LOFMAX.index] = lofmax
	astats[__statsorder.LOFMEAN.index] = lofmean
	astats[__statsorder.LOFSTD.index] = lofstd
	(scode, deriv, desc, het, hom, hethom) = \
			_distinguishFeatures(afeatures)
	astats[__statsorder.SCODE.index] = len(scode.keys())
	astats[__statsorder.DERIV.index] = len(deriv.keys())
	astats[__statsorder.DESC.index] = len(desc.keys())
	astats[__statsorder.HET.index] = len(het.keys())
	astats[__statsorder.HOM.index] = len(hom.keys())
	astats[__statsorder.HOHE.index] = len(hethom.keys())
	astats[__statsorder.SDEGMEAN.index] = sdegmean
	astats[__statsorder.SDEGSTD.index] = sdegstd
	astats[__statsorder.TDEGMEAN.index] = tdegmean
	astats[__statsorder.TDEGSTD.index] = tdegstd

	# write the feature expression for the whole project
	# take the sigmap and join all expressions by '&&'
	#exprs = map(lambda n: n[0], sigmap.itervalues())
	exprs = map(lambda n: n, sigmap.iterkeys())
	#featureexpr = '&&'.join(exprs)
	featureexpr = '&'.join(exprs)
	fdfeat.write(featureexpr)
	fdfeat.close()

	fdcsv.writerow(astats)
	fd.close()
	if options.maple:
		print('feature expressions: not equal '
			'string to maple (%s)' % str(__errorfexp))
		print('feature expressions: found error '
			'in (%s) expressions' % str(len(__errormatch)))
		print(__errormatch)


##################################################
if __name__ == '__main__':

	folder = os.path.abspath(options.folder)
	if (os.path.isdir(folder)):
		apply(folder)
	else:
		sys.exit(-1)
